/* ************************************************************************** 
   Matlab mex wrapper for the embed function in the Planar Inference code of Nic Schraudolph
   -- Dhruv Batra (batradhruv -at- cmu.edu) 
   ************************************************************************** */

#include "mex.h"
#include "matrix.h"

// Let Matlab handle memory allocation
#define malloc mxMalloc
#define calloc mxCalloc 
#define free mxFree
#define realloc mxRealloc
#define printf mexPrintf
//#define printf mexErrMsgTxt

// row-major array access
#define ARR_RM(arr, r_ind, c_ind, ncols) (*(arr + r_ind*ncols + c_ind))
// col-major array access
#define ARR_CM(arr, r_ind, c_ind, nrows) (*(arr + c_ind*nrows + r_ind))


#include "src/utils.hpp"
#include "planarity/graph.h"

#include <fstream>
#include <string>
#include <map>

typedef std::pair<size_t,size_t> Edge;

// set up map from node pair to edge index
typedef std::pair<Edge, size_t> EdgeMapEntry;
typedef std::map<Edge, size_t> EdgeMap;
EdgeMap edgemap;

// handle planarity return codes
inline void errCheck(int k)
{
  if (k == OK) return;

  switch (k)
    {
    case NOTOK:
      edgemap.clear();
      mexErrMsgTxt("something is wrong\n");
      break;
    case NONPLANAR:
      edgemap.clear();
      mexErrMsgTxt("graph is not planar\n");
      break;
    case NONOUTERPLANAR:
      edgemap.clear();
      mexErrMsgTxt("graph is not outerplanar\n");
      break;
    }
}

// write out adjacency list of edge indices
void copyAdjList(graphP g, double *xvalues, int max_degree)
{
  int i, j, counter;
  char s[50];

  for (i = 0; i < g->N; ++i)
    {
      counter = 0;
      j = g->G[i].link[1];
      while (j >= g->N)
	{
	  Edge e(std::min(i, g->G[j].v), std::max(i, g->G[j].v));

	  EdgeMap::iterator it(edgemap.find(e));
	  if (it == edgemap.end()) {
	    sprintf(s,"can't find edge (%d,%d)\n",i,g->G[j].v);
	    edgemap.clear();
	    mexErrMsgTxt(s);
	  }
	    
	  //std::cout << it->second << ' ';
	  ARR_CM(xvalues,i,counter,g->N) = (double) it->second;
	  counter++;

	  j = g->G[j].link[1];
	}
      
      //std::cout << std::endl;
      while (counter < max_degree) {
	ARR_CM(xvalues,i,counter,g->N) = -1;
	counter++;
      }
      
    }
}

// write out adjacency list of edge indices
void writeAdjList(graphP g)
{
     int i, j;

     for (i = 0; i < g->N; ++i)
     {
          j = g->G[i].link[1];
          while (j >= g->N)
          {
              Edge e(std::min(i, g->G[j].v), std::max(i, g->G[j].v));

              EdgeMap::iterator it(edgemap.find(e));
              if (it == edgemap.end())
		{
                  std::cerr << Isinf::argv0 << "can't find edge ("
                      << i << ',' << g->G[j].v << ')' << std::endl;
                  exit(5);
              }
              std::cout << it->second << ' ';

              j = g->G[j].link[1];
          }
          std::cout << std::endl;
     }
}


/* ************************************************************************** 
   mex Wrapper
   ************************************************************************** */

/* Function: [embedding] = embed(edgelist)
   ----------------------
   Inputs:
   edgelist       <num_edges> x 2 adjacency list matrix. Each row is a pair of vertices 
   corresponding to that edge. Nodes are edges are indexed by integers (starting from 0).

   Outputs:
   embedding      <num_nodes> x Delta sparse embedding matrix, where Delta is the max-degree of all nodes. 
                  Each row indicates an orientation of the edges incident to that node. 
   ----------------------
*/

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  
  int n_nodes,n_edges;
  int row,col,i,j;
  mxArray *xdata;
  double *xvalues;
  int max_degree=1;
  char s[50];
  int v1,v2;

  if (nrhs != 1) 
    mexErrMsgTxt("Incorrect number of input arguments!");

  if (nlhs != 1) 
    mexErrMsgTxt("Out of range: number of output arguments!");
  
  
  /* READ INPUT -> 'edgelist' which is a n_edges X 2 matrix */
  xdata = (mxArray *)prhs[0];
  row = mxGetM(xdata);
  col = mxGetN(xdata);
  
  n_edges = row;

  if ((col != 2) || (n_edges <= 0)) 
    mexErrMsgTxt("Error: Incorrect size of edgeslist (input arg #1)!");
  
  xvalues = mxGetPr(xdata);
  /* READ DONE */


  std::vector<Edge> edges;
  //std::string s1,s2;
  size_t max = 0;
  int n = 0;
  
  for (i=0; i<n_edges; i++)
    {
      v1 = (int)ARR_CM(xvalues,i,0,n_edges);
      v2 = (int)ARR_CM(xvalues,i,1,n_edges);

      if (v1 == v2) { // self-loop! 
	edgemap.clear();
	sprintf(s,"error at node %d: self-loops not allowed\n",v1);
	mexErrMsgTxt(s);
      }
      
      if (v1 > v2) std::swap(v1, v2);
      if (v2 > max) max = (size_t) v2;

      Edge e(v1, v2);

      edges.push_back(e);

      if (!edgemap.insert(EdgeMapEntry(e, n++)).second) {
	edgemap.clear();
	sprintf(s,"duplicate edge between nodes %d and %d\n", v1, v2);
	mexErrMsgTxt(s);
      }
    }
  if (edges.size() < 2){      
    edgemap.clear();
    mexErrMsgTxt("less than 2 valid edges found\n");
  }
  else if (edges.size() > 3*(++max) - 6) {
    // edge limit for planar graphs
    errCheck(NONPLANAR);
  }

  // create graph in planarity code
  n_nodes = max;
  graphP g = gp_New();
  gp_InitGraph(g, max);
  for (std::vector<Edge>::iterator i = edges.begin(); i < edges.end(); ++i)
    errCheck(gp_AddEdge(g, i->first, 0, i->second, 0));
  
  //writeAdjList(g);
  // timed call to planarity code
  Isinf::CPUtimer *t = new Isinf::CPUtimer;
  n = gp_Embed(g, EMBEDFLAGS_PLANAR);
  delete t;

  // report result
  if (n != OK)
    {
      gp_Free(&g);
      errCheck(n);
    }
  
  //printf("n_edges = %d, max = %d\n",edges.size(),max);
  gp_SortVertices(g);
  
  // get max-degree from graph 
  for (i=0; i<n_nodes; i++) {
    j = gp_GetVertexDegree(g,i);
    if (max_degree < j)
      max_degree = j;
  }
  
  // create space for output variables
  plhs[0] = mxCreateDoubleMatrix(n_nodes, max_degree, mxREAL);
  xvalues = mxGetPr(plhs[0]);
  
  copyAdjList(g,xvalues,max_degree);  // output planar embedding
  
  //free variables
  gp_Free(&g);
  edgemap.clear();

  return;
  
}



