/* ************************************************************************** 
   Matlab mex wrapper for the map function in the Planar Inference code of Nic Schraudolph
   -- Dhruv Batra (batradhruv -at- cmu.edu) 
   ************************************************************************** */

#include "mex.h"
#include "matrix.h"

// Let Matlab handle memory allocation
#define malloc mxMalloc
#define calloc mxCalloc 
#define free mxFree
#define realloc mxRealloc
#define printf mexPrintf
#define assert mxAssert
#define error mexErrMsgTxt


// row-major array access
#define ARR_RM(arr, r_ind, c_ind, ncols) (*(arr + r_ind*ncols + c_ind))
// col-major array access
#define ARR_CM(arr, r_ind, c_ind, nrows) (*(arr + c_ind*nrows + r_ind))

#include "blossom_mex.hpp"


/* ************************************************************************** 
   mex Wrapper
   ************************************************************************** */

/* Function: [nodestates, lb, ub] = map(embedding,edgefeatures,params)
   ----------------------
   Inputs:
   embedding      <num_nodes> x <max_degree> embedding matrix.
                  See embed wrapper for more details.
   edgefeatures   <num_edges> x <num_feat> matrix. Each row hold the feature vector for that edge.
   params         1 x <num_feat> row vector of weights for features.
		  
   Outputs:
   nodestates     <num_nodes> x 1 vector of MAP states for the nodes.
   lb             scalar holding lower bound on energy of MAP state.
   ub             scalar holding upper bound on energy of MAP state.
   ----------------------
*/

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  int n_nodes,n_edges,max_degree,n_feat;
  int row,col,i,j;
  mxArray *xdata;
  double *xvalues,*pr;
  int *ir,*jc;
  double w;

  if (nrhs != 3) 
    mexErrMsgTxt("Incorrect number of input arguments!");
  
  if (nlhs < 1) 
    mexErrMsgTxt("Incorrect number of output arguments!");
  
  //if (!mxIsClass(prhs[0],"sparse"))
  //mexErrMsgTxt("embedding must be a sparse matrix (see help)!");

  /* READ 1st INPUT -> 'embedding' which is a <n_nodes> x <max_degree> matrix */
  xdata = (mxArray *)prhs[0];
  row = mxGetM(xdata);
  col = mxGetN(xdata);
  
  n_nodes = row;
  max_degree = col;


  if ((max_degree <=0) || (n_nodes <= 0)) 
    mexErrMsgTxt("Error: Incorrect size of embedding (input arg #1)!");
  
  xvalues = mxGetPr(xdata);
  /* READ DONE */
  
  // create graph from first argument
  Isinf::mexGraph g(xvalues,n_nodes,max_degree);
  assert(g.is_connected());
  if (g.genus() > 0)
    printf("warning: graph has genus %d (Not planar)\n",g.genus());

  n_edges = g.nedges;
  size_t n = g.angulate();

  
  //g.adj_list(std::cout);

  /* READ 3rd INPUT -> 'params' which is a 1 x <num_feat> vector */
  xdata = (mxArray *)prhs[2];
  row = mxGetM(xdata);
  col = mxGetN(xdata);
  
  if ((row>1) && (col>1))
    mexErrMsgTxt("params (input arg #3) must be a vector!");
  
  n_feat = (row > col) ? row : col;
  xvalues = mxGetPr(xdata);
  /* READ DONE */
  
  /* READ 2rd INPUT -> 'edgefeatures' which is a <num_edges> x <num_feat> matrix */
  xdata = (mxArray *)prhs[1];
  row = mxGetM(xdata);
  col = mxGetN(xdata);
  
  if ((row != n_edges) || (col != n_feat)) 
    mexErrMsgTxt("Incorrect size of edgefeatures (input arg #2)!");
  
  pr = mxGetPr(xdata);
  /* READ DONE */
  
  std::vector<double> e;
  e.reserve(n_edges);

  for (i=0; i<n_edges; i++) {
    w = 0;
    for (j=0; j<n_feat; j++) 
      w += ARR_CM(pr,i,j,n_edges)*xvalues[j];
    e.push_back(w);
  }

  // find extended ground edge & MMC node state
  Isinf::CPUtimer *t = new Isinf::CPUtimer;
  g.call_blossom(e);
  
  std::vector<double> slacks;
  slacks.reserve(g.nedges);
  for (size_t i = 0; i < g.nedges; ++i)
    slacks.push_back(g.pm->ExtendedSlack(i));
  g.mmc_nodestate(slacks);
  delete t;

  // calculate & compare energies
  double lb = 0.0;
  double ub = 0.0;
  for (size_t i = 0; i < e.size(); ++i)
    {
      if (!g.edges[i]->agree) lb += e[i];
      if (g.nodes[g.edges[i]->node].state !=
	  g.nodes[g.edges[i]->other->node].state) ub += e[i];
    }
  bool same = (std::abs(lb-ub) <= std::abs(lb)*1e-12);
  if (!g.genus()) assert(same);
  else assert(same || lb <= ub);

  // try quenching
  if (!same)
    {
      // std::cerr << std::setprecision(12) << ub << std::endl;
      ub += g.edge_quench(e);
      // std::cerr << ub << std::endl;
      ub += g.node_quench(e);
      // std::cerr << ub << std::endl;
    }


  /* Prepare space for output */
  plhs[0] = mxCreateDoubleMatrix(n_nodes, 1, mxREAL);
  xvalues = mxGetPr(plhs[0]);
  
  if (nlhs > 1)
    plhs[1] = mxCreateDoubleScalar(lb);
  if (nlhs > 2)
    plhs[2] = mxCreateDoubleScalar(ub);
  

  // report node state & energy
  for (i = 0; i < g.nodes.size(); ++i)
    xvalues[i] = (g.nodes[i].state != g.nodes.back().state ? 1 : 0);

  e.clear();
}
