/* ************************************************************************** 
   Matlab mex wrapper that computes min-marginals via Dynamic Planar Cuts. This utilizes
   the Planar Inference code of Nic Schraudolph, and Bloosom V code of Vladimir Kolmogorov.
   
   Copyright 2009 Dhruv Batra (batradhruv -at- cmu.edu)
   ************************************************************************** */


#include "mex.h"
#include "matrix.h"

// Let Matlab handle memory allocation
#define malloc mxMalloc
#define calloc mxCalloc 
#define free mxFree
#define realloc mxRealloc
#define printf mexPrintf
#define assert mxAssert
#define error mexErrMsgTxt


// row-major array access
#define ARR_RM(arr, r_ind, c_ind, ncols) (*(arr + r_ind*ncols + c_ind))
// col-major array access
#define ARR_CM(arr, r_ind, c_ind, nrows) (*(arr + c_ind*nrows + r_ind))

#include "blossom_mex.hpp"


/* ************************************************************************** 
   mex Wrapper
   ************************************************************************** */

/* Function: [min_marg] = min_marg_mex(embedding,edge_weights,delta_inds, verbose_level<optional>)
   ----------------------
   Inputs:
   embedding      <num_nodes> x <max_degree> embedding matrix.
                  See embed wrapper for more details.
   edge_weights   <num_edges> x 1 vector holding the weights on edges.
   delta_inds     <num_nodes-1> x 1 vector holding indices of s-i edges for non-source inds i. 0-based. Uint32
   verbose_level  scalar: 0 for no messages (DEFAULT); 1 for some.

   Outputs:
   min_marg       2 x <num_nodes-1> matrix holding min-marginals for all non-source nodes.
   ----------------------
*/


void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  int n_nodes,n_edges,max_degree;
  int row,col,i,j;
  double *xvalues;
  double *ew2;
  uint32_t *di;

  if ( (nrhs!=3) && (nrhs!=4) )
    mexErrMsgTxt("Incorrect number of input arguments!: Must be 3 or 4");
  
  if (nlhs > 1) 
    mexErrMsgTxt("Incorrect number of output arguments!: Must be at most 1");
  
  /* READ 1st INPUT -> 'embedding' */
  n_nodes = mxGetM(prhs[0]);
  max_degree = mxGetN(prhs[0]);
  
  if ((max_degree <=0) || (n_nodes <= 0)) 
    mexErrMsgTxt("Error: Incorrect size of embedding (input arg #1)!");
  
  xvalues = mxGetPr(prhs[0]);

  // create graph from first argument
  Isinf::mexGraph g(xvalues,n_nodes,max_degree);
  assert(g.is_connected());
  if (g.genus() > 0)
    printf("warning: graph has genus %d (Not planar)\n",g.genus());

  n_edges = g.nedges;
  size_t n = g.angulate();
  /* READ DONE */

  //g.adj_list(std::cout);

  /* READ 2rd INPUT -> 'edge_weights' */
  row = mxGetM(prhs[1]);
  col = mxGetN(prhs[1]);
  
  if ((row>1) && (col>1))
    mexErrMsgTxt("edge_weights (input arg #2) must be a vector!");
  
  if ( n_edges != ((row > col) ? row : col) )
    error("Incorrect size of edge_weights!");

  ew2 = mxGetPr(prhs[1]); // Because ew will be changed.
  std::vector<double> ew(n_edges,0);
  memcpy(&ew[0],mxGetPr(prhs[1]),n_edges*sizeof(double));
  /* READ DONE */
  
  /* READ 3rd INPUT -> 'delta_inds' */
  row = mxGetM(prhs[2]);
  col = mxGetN(prhs[2]);
  
  if ((row>1) && (col>1))
    error("delta_inds (input arg #4) must be a vector!");
  
  if ( (n_nodes-1) != ( (row > col) ? row : col ) )
    error("Incorrect size of delta_inds!");

  if (!mxIsUint32(prhs[2]))
    error("delta_inds should be uint32");
  di = (uint32_t*) mxGetPr(prhs[2]);
  /* READ DONE */

  /* READ 4th INPUT -> 'verbose_level' */
  if (nrhs>3)
    g.verbose_level = (int) mxGetScalar(prhs[3]);
  /* READ DONE */

  /* Prepare space for output */
  plhs[0] = mxCreateDoubleMatrix(2, n_nodes-1, mxREAL);

  // find MAP edge state
  g.blossom_minmarg(mxGetPr(plhs[0]),ew, di);
  
}
